

/*
 *    Stopwords.java
 *    Copyright (C) 2001 Eibe Frank
 */

//package weka.core;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.Collections;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

/**
 * Class that can test whether a given string is a stop word.
 * Lowercases all words before the test. <p/>
 * The format for reading and writing is one word per line, lines starting
 * with '#' are interpreted as comments and therefore skipped. <p/>
 * The default stopwords are based on <a href="http://www.cs.cmu.edu/~mccallum/bow/rainbow/" target="_blank">Rainbow</a>. <p/>
 *
 * Accepts the following parameter: <p/>
 *
 * -i file <br/>
 * loads the stopwords from the given file <p/>
 *
 * -o file <br/>
 * saves the stopwords to the given file <p/>
 *
 * -p <br/>
 * outputs the current stopwords on stdout <p/>
 *
 * Any additional parameters are interpreted as words to test as stopwords.
 * 
 * @author Eibe Frank (eibe@cs.waikato.ac.nz)
 * @author Ashraf M. Kibriya (amk14@cs.waikato.ac.nz)
 * @author FracPete (fracpete at waikato dot ac dot nz)
 * @version $Revision: 1.4 $
 */
public class Stopwords {
  
  /** The hash set containing the list of stopwords */
  protected HashSet m_Words = null;

  /** The default stopwords object (stoplist based on Rainbow) */
  protected static Stopwords m_Stopwords;

  static {
    if (m_Stopwords == null) {
      m_Stopwords = new Stopwords();
    }
  }

  /**
   * initializes the stopwords (based on <a href="http://www.cs.cmu.edu/~mccallum/bow/rainbow/" target="_blank">Rainbow</a>).

   */
  
  public Stopwords() {
    m_Words = new HashSet();

    //Stopwords list from Rainbow
    add("a");
    add("able");
    add("about");
    add("above");
    add("according");
    add("accordingly");
    add("across");
    add("actually");
    add("after");
    add("afterwards");
    add("again");
    add("against");
    add("all");
    add("allow");
    add("allows");
    add("allowing");
    add("allowed");
    add("almost");
    add("alone");
    add("along");
    add("already");
    add("also");
    add("although");
    add("always");
    add("am");
    add("among");
    add("amongst");
    add("an");
    add("and");
    add("another");
    add("any");
    add("anybody");
    add("anyhow");
    add("anymore");
    add("anyone");
    add("anything");
    add("anyway");
    add("anyways");
    add("anywhere");
    add("apart");
    add("appear");
//    add("appreciate");
    add("appropriate");
    add("are");
    add("around");
    add("as");
    add("aside");
    add("ask");
    add("asking");
    add("associated");
    add("at");
    add("available");
    add("away");
//    add("awfully");
    add("b");
    add("be");
    add("became");
    add("because");
    add("become");
    add("becomes");
    add("becoming");
    add("been");
    add("before");
    add("beforehand");
    add("behind");
    add("being");
    add("believe");
    add("below");
    add("beside");
    add("besides");
//    add("best");
//    add("better");
    add("between");
    add("beyond");
    add("both");
    add("but");    
    add("brief");
    add("by");
    add("c");
    add("came");
    add("can");
    add("certain");
    add("certainly");
    add("clearly");
    add("co");
    add("com");
    add("come");
    add("comes");
    add("contain");
    add("containing");
    add("contains");
    add("corresponding");
    add("could");
    add("course");
    add("currently");
    add("d");
    add("definitely");
    add("described");
    add("despite");
    add("did");
    add("different");
    add("do");
    add("does");
    add("doing");
    add("done");
    add("down");
    add("downwards");
    add("during");
    add("e");
    add("each");
    add("edu");
    add("eg");
    add("eight");
    add("either");
    add("else");
    add("elsewhere");
    add("enough");
    add("entirely");
    add("especially");
    add("et");
    add("etc");
    add("even");
    add("ever");
    add("every");
    add("everybody");
    add("everyone");
    add("everything");
    add("everywhere");
    add("eventually");
    add("ex");
    add("exactly");
    add("example");
    add("except");
    add("f");
    add("far");
    add("few");
    add("fifth");
    add("first");
    add("five");
    add("followed");
    add("following");
    add("follows");
    add("for");
    add("former");
    add("formerly");
    add("forth");
    add("four");
    add("from");
    add("further");
    add("furthermore");
    add("g");
    add("get");
    add("gets");
    add("getting");
    add("given");
    add("gives");
    add("go");
    add("goes");
    add("going");
    add("gone");
    add("got");
    add("gotten");
    add("greetings");
    add("h");
    add("had");
    add("happens");
    add("hardly");
    add("has");
    add("have");
    add("having");
    add("he");
    add("hello");
    add("help");
    add("hence");
    add("her");
    add("here");
    add("hereafter");
    add("hereby");
    add("herein");
    add("hereupon");
    add("hers");
    add("herself");
    add("hi");
    add("him");
    add("himself");
    add("his");
    add("hither");
    add("hopefully");
    add("how");
    add("howbeit");
    add("however");
    add("i");
    add("ie");
    add("if");
//    add("ignored");
    add("immediate");
    add("in");
    add("inasmuch");
    add("inc");
    add("indeed");
    add("indicate");
    add("indicated");
    add("indicates");
    add("inner");
    add("insofar");
    add("instead");
    add("into");
    add("inward");
    add("is");
    add("it");
    add("its");
    add("itself");
    add("j");
    add("just");
    add("k");
    add("keep");
    add("keeps");
    add("kept");
//    add("know");
//    add("knows");
//    add("known");
    add("l");
    add("last");
    add("lately");
    add("later");
    add("latter");
    add("latterly");
    add("least");
    add("less");
    add("lest");
    add("let");
    add("like");
    add("liked");
    add("likely");
    add("little");
    add("ll"); //added to avoid words like you'll,I'll etc.
    add("look");
    add("looking");
    add("looks");
    add("ltd");
    add("m");
    add("mainly");
    add("many");
    add("may");
    add("maybe");
    add("me");
//    add("mean");
    add("meanwhile");
//    add("merely");
    add("might");
    add("more");
    add("moreover");
    add("most");
    add("mostly");
    add("much");
    add("must");
    add("my");
    add("myself");
    add("n");
    add("name");
    add("namely");
    add("nd");
    add("near");
    add("nearly");
    add("necessary");
    add("need");
    add("needs");
    add("neither");
//    add("never");
    add("nevertheless");
    add("new");
    add("next");
    add("nine");
    add("normally");
//    add("novel");
    add("no");
    add("nobody");
    add("non");
    add("none");
    add("noone");
    add("nor");
    add("normally");
    add("not");
    add("n't");
    add("nothing");
    add("novel");
    add("now");
    add("nowhere");
    add("now");
    add("nowhere");
    add("o");
    add("obviously");
    add("of");
    add("off");
    add("often");
    add("oh");
    add("ok");
    add("okay");
//    add("old");
    add("on");
    add("once");
    add("one");
    add("ones");
    add("only");
    add("onto");
    add("or");
    add("other");
    add("others");
    add("otherwise");
    add("ought");
    add("our");
    add("ours");
    add("ourselves");
    add("out");
    add("outside");
    add("over");
    add("overall");
    add("own");
    add("p");
    add("particular");
    add("particularly");
    add("per");
    add("perhaps");
    add("placed");
    add("please");
    add("plus");
    add("possible");
    add("presumably");
    add("probably");
    add("provides");
    add("q");
    add("que");
    add("quite");
    add("qv");
    add("r");
    add("rather");
    add("rd");
    add("re");
    add("really");
    add("reasonably");
    add("regarding");
    add("regardless");
    add("regards");
    add("relatively");
    add("respectively");
    add("right");
    add("s");
    add("said");
    add("same");
    add("saw");
    add("say");
    add("saying");
    add("says");
    add("second");
    add("secondly");
    add("see");
    add("seeing");
    add("seem");
    add("seemed");
    add("seeming");
    add("seems");
    add("seen");
    add("self");
    add("selves");
    add("sensible");
    add("sent");
   // add("serious");
   // add("seriously");
    add("seven");
    add("several");
    add("shall");
    add("she");
    add("should");
    add("since");
    add("six");
    add("so");
    add("some");
    add("somebody");
    add("somehow");
    add("someone");
    add("something");
    add("sometime");
    add("sometimes");
    add("somewhat");
    add("somewhere");
    add("soon");
    add("sorry");
    add("specified");
    add("specify");
    add("specifying");
    add("still");
    add("sub");
    add("such");
    add("sup");
    add("sure");
    add("t");
    add("take");
    add("taken");
    add("tell");
    add("tends");
    add("th");
    add("than");
    add("thank");
    add("thanks");
    add("thanx");
    add("that");
    add("thats");
    add("the");
    add("their");
    add("theirs");
    add("them");
    add("themselves");
    add("then");
    add("thence");
    add("there");
    add("thereafter");
    add("thereby");
    add("therefore");
    add("therein");
    add("theres");
    add("thereupon");
    add("these");
    add("they");
    add("think");
    add("third");
    add("this");
    add("thorough");
    add("thoroughly");
    add("those");
    add("though");
    add("three");
    add("through");
    add("throughout");
    add("thru");
    add("thus");
    add("till");
    add("to");
    add("together");
    add("too");
    add("took");
    add("toward");
    add("towards");
    add("tried");
    add("tries");
    add("truly");
    add("try");
    add("trying");
    add("twice");
    add("two");
    add("u");
    add("un");
    add("under");
//    add("unfortunately");
    add("unless");
    add("unlikely");
    add("until");
    add("unto");
    add("up");
    add("upon");
    add("us");
    add("use");
    add("used");
//    add("useful");
    add("uses");
    add("using");
    add("usually");
    add("uucp");
    add("v");
    add("value");
    add("various");
    add("ve"); //added to avoid words like I've,you've etc.
    add("very");
    add("via");
    add("viz");
    add("vs");
    add("w");
    add("want");
    add("wants");
    add("was");
//    add("way");
    add("we");
//    add("welcome");
    add("well");
    add("went");
    add("were");
    add("what");
    add("whatever");
    add("when");
    add("whence");
    add("whenever");
    add("where");
    add("whereafter");
    add("whereas");
    add("whereby");
    add("wherein");
    add("whereupon");
    add("wherever");
    add("whether");
    add("which");
    add("while");
    add("whither");
    add("who");
    add("whoever");
    add("whole");
    add("whom");
    add("whose");
    add("why");
    add("will");
    add("willing");
    add("wish");
    add("with");
    add("within");
    add("without");
    add("wonder");
    add("would");
    add("would");
    add("x");
    add("y");
//    add("yes");
    add("yet");
    add("you");
    add("your");
    add("yours");
    add("yourself");
    add("yourselves");
    add("z");
    add("zero");
    // add new
    add("i'm");
    add("he's");
    add("she's");
    add("you're");
    add("i'll");
    add("you'll");
    add("she'll");
    add("he'll");
    add("it's");
    add("don't");
    add("can't");
    add("didn't");
    add("i've");
    add("that's");
    add("there's");
    add("isn't");
    add("what's");
    add("rt");
    add("doesn't");
    add("w/");
    add("w/o");
    /**the following add are all java keywords*/
    /*
    add("abstract");
    add("assert");
    add("boolean");
    add("break");
    add("byte");
    add("case");
    add("catch");
    add("char");
    add("class");
    add("const");
    add("continue");
    add("default");
    add("do");
    add("double");
    add("else");
    add("enum");
    add("extends");
    add("final");
    add("finally");
    add("float");
    add("for");
    add("goto");
    add("if");
    add("implements");
    add("import");
    add("instanceof");
    add("int");
    add("interface");
    add("long");
    add("native");
    add("new");
    add("null");
    add("package");
    add("private");
    add("protected");
    add("public");
    add("return");
    add("short");
    add("static");
    add("strictfp");
    add("super");
    add("switch");
    add("synchronized");
    add("this");
    add("throw");
    add("throws");
    add("transient");
    add("try");
    add("void");
    add("volatile");
    add("while");
    add("java");
    add("string");
    //*/
    add("isn");
    add("ll");
    add("ve");
    add("didn");
    add("doesn");
    add("sourceforge");
    add("sf");
    add("don");
    
    add("wasn");
    add("www");
    add("oops");
    add("haven");
    add("hasn");
    add("aren");
    add("couldn");
    add("wouldn");
    add("whatever");
    add("awhile");
    add("weren");
    add("hadn");
    //the following words are all people names.
    add("tatsuo");
    add("john");
    add("gray");
    add("bernhard");
    add("herzog");
    add("fernando");
    add("nasser");
    add("dave");
    add("dominic");
    add("eidson");
    add("ishii");
    add("josh");
    add("berkus");
    add("barry");
    add("gavin");
    add("sherry");
    add("lind");
    add("viktor");
    add("vislobokov");
    add("eisentraut");
    add("oleg");
    add("chris");
    add("alvaro");
    add("andrew");
    add("andreas");
    add("dunstan");
    add("michael");
    add("tom");
    add("gavin");
    add("rod");
    add("taylor");
    add("kings");
    add("lynne");
    add("christopher");
    add("marko");
    add("bruce");
    add("philip");
    add("sullivan");
    add("tom");
    add("conway");
    add("conwai");
    add("greg");
    add("sabino");
    add("mullane");
    add("marcus");
    add("kris");
    add("jurka");
    add("giles");
    add("bear");
    add("simon");
    add("riggs");
    add("teodor");
    add("sigaev");
    add("david");
    add("neil");
    add("alvaro");
    add("herrera");
    add("kreen");
    add("hagander");
    add("magnus");
    add("naeslund");
    add("joe");
    add("takahiro");
    add("itagki");
    add("hiroshi");
    add("saito");
    add("thomas");
    add("thai");
    add("minnesota");
    add("teodor");
    add("sigaev");
    add("peter");
    add("alexey");
    add("borzov");
    add("halley");
    add("pacheco");
    add("oliveira");
    add("de");
    add("euler");
    add("taveira");
    add("warner");
    add("martijn");
    add("van");
    add("ooster");
    add("aasmund");
    add("midttun");
    add("godal");
    add("karel");
    add("zak");
    add("vadim");
    add("lee");
    add("oliver");
    add("elphick");
    add("jane");
    add("marc");
    add("fournier");
    add("brent");
    add("verner");
    add("japanese");
    add("jun");
    add("kuwamura");
    add("ian");
    add("barwick");
    add("meskes");
    add("jason");
    add("tishler");
    add("lane");
    add("ryouichi");
    add("matsuda");
    add("matuda");
    add("sra");
    add("jp");
    add("joachim");
    add("wieland");
    add("robert");
    add("bruccoleri");
    add("pavan");
    add("deolasee");
    add("guillaume");
    add("lelarge");
    add("zdenek");
    add("kotala");
    add("laurenz");
    add("albe");
    add("joshua");
    add("drake");
    add("volkan");
    add("yazici");
    add("itagaki");
    add("sergey");
    add("koposov");
    add("atsushi");
    add("ogawa");
    add("heikki");
    add("linnakangas");
    add("brendan");
    add("jurd");
    add("jaime");
    add("casanova");
    add("stefan");
    add("kaltenbrunner");
    add("andreas");
    add("seltenreich");
    add("bernd");
    add("helmle");
    add("lelarge");
    add("guillaume");
    add("abhijit");
    add("menon");
    add("sen");
    add("merlin");
    add("moncure");
    add("elphick");
    add("tokuya");
    add("rbt");
    add("ca");
    add("ru");
    add("stephan");
    add("szabo");
    add("bartunov");
    add("oleg");
    add("em");
    add("florian");
    add("oliver");
    add("elphick");
    add("jim");
    add("mauro");
    add("richard");
    add("mcdougall");
    add("josh");
    add("rovero");
    add("au");
    add("nconwai");
    add("manfred");
    add("koizar");
    add("giles");
    add("bear");
    add("william");
    add("volkman");
    add("mike");
    add("ron");
    add("snyder");
    add("ooops");
    add("martijn");
    add("van");
    add("oosterhout");
    add("dmitry");
    add("tkach");
    add("steven");
    add("singer");
    add("manfred");
    add("koizar");
    add("zhou");
    add("han");
    add("dennis");
    add("bjorklund");
    add("sergey");
    add("smirnov");
    add("martin");
    add("renters");
    add("bruno");
    add("wolff");
    add("nigel");
    add("andrews");
    add("iii");
    add("sean");
    add("chittenden");
    add("stephen");
    add("marshall");
    add("brendan");
    add("lefebvre");
    add("mea");
    add("culpa");
    add("larry");
    add("rosenman");
    add("harris");
    add("jeff");
    add("davis");
    add("wheeler");
    add("momjian");
    add("lester");
    add("godwin");
    add("jowett");
    add("matthew");
    add("vanecek");
    add("greg");
    add("stark");
    add("pavel");
    add("hanak");
    add("kevin");
    add("brown");
    add("haas");
    add("connor");
    add("matthew");
    add("moi");
    add("fabien");
    add("jeremy");
    add("yoder");
    add("hallgren");
    add("glaesemann");
    add("fetter");
    add("fuhr");
    add("mayer");
    add("pitt");
    add("steve");
    add("wolfe");
    add("devrim");
    add("kurt");
    add("roeckx");
    add("nasby");
    add("kim");
    add("ho");
    add("markus");
    add("bertheau");
    add("mark");
    add("kirkwood");
    add("smith");
    add("brian");
    add("gallew");
    add("claudio");
    add("natoli");
    add("stehule");
    add("coelho");
    add("james");
    
    add("http");
    add("org");
    add("msg");
    add("php");
    add("archives");
    add("thank");
    add("thanks");
    add("pg");
    add("pa");
    add("postgresql");
    add("imho");//ˡֱ
    add("kibitzing");//ҳע⣬
    add("pha");
    add("pgsql");
    add("pl");
    add("psql");
    add("hacker");
    add("hackers");
    add("postgres");
    //*/
  }

  
  /**
   * removes all stopwords
   */
  public void clear() {
    m_Words.clear();
  }

  /**
   * adds the given word to the stopword list (is automatically converted to
   * lower case and trimmed)
   *
   * @param word the word to add
   */
  public void add(String word) {
    if (word.trim().length() > 0)
      m_Words.add(word.trim().toLowerCase());
  }

  /**
   * removes the word from the stopword list
   *
   * @param word the word to remove
   * @return true if the word was found in the list and then removed
   */
  public boolean remove(String word) {
    return m_Words.remove(word);
  }
  
  /**
   * Returns a sorted enumeration over all stored stopwords
   *
   * @return the enumeration over all stopwords
   */
  public Enumeration elements() {
    Iterator    iter;
    Vector      list;

    iter = m_Words.iterator();
    list = new Vector();

    while (iter.hasNext())
      list.add(iter.next());

    // sort list
    Collections.sort(list);
    return list.elements();
  }

  /**
   * Generates a new Stopwords object from the given file
   *
   * @param filename the file to read the stopwords from
   * @throws Exception if reading fails
   */
  public void read(String filename) throws Exception {
    read(new File(filename));
  }

  /**
   * Generates a new Stopwords object from the given file
   *
   * @param file the file to read the stopwords from
   * @throws Exception if reading fails
   */
  public void read(File file) throws Exception {
    read(new BufferedReader(new FileReader(file)));
  }

  /**
   * Generates a new Stopwords object from the reader. The reader is
   * closed automatically.
   *
   * @param reader the reader to get the stopwords from
   * @throws Exception if reading fails
   */
  public void read(BufferedReader reader) throws Exception {
    String      line;

    clear();
    
    while ((line = reader.readLine()) != null) {
      line = line.trim();
      // comment?
      if (line.startsWith("#"))
        continue;
      add(line);
    }

    reader.close();
  }

  /**
   * Writes the current stopwords to the given file
   *
   * @param filename the file to write the stopwords to
   * @throws Exception if writing fails
   */
  public void write(String filename) throws Exception {
    write(new File(filename));
  }

  /**
   * Writes the current stopwords to the given file
   *
   * @param file the file to write the stopwords to
   * @throws Exception if writing fails
   */
  public void write(File file) throws Exception {
    write(new BufferedWriter(new FileWriter(file)));
  }

  /**
   * Writes the current stopwords to the given writer. The writer is closed
   * automatically.
   *
   * @param writer the writer to get the stopwords from
   * @throws Exception if writing fails
   */
  public void write(BufferedWriter writer) throws Exception {
    Enumeration   enm;

    // header
    writer.write("# generated " + new Date());
    writer.newLine();

    enm = elements();

    while (enm.hasMoreElements()) {
      writer.write(enm.nextElement().toString());
      writer.newLine();
    }

    writer.flush();
    writer.close();
  }

  /**
   * returns the current stopwords in a string
   *
   * @return the current stopwords
   */
  public String toString() {
    Enumeration   enm;
    StringBuffer  result;

    result = new StringBuffer();
    enm    = elements();
    while (enm.hasMoreElements()) {
      result.append(enm.nextElement().toString());
      if (enm.hasMoreElements())
        result.append(",");
    }

    return result.toString();
  }
  
  /** 
   * Returns true if the given string is a stop word.
   * 
   * @param word the word to test
   * @return true if the word is a stopword
   */
  public boolean is(String word) {
    return m_Words.contains(word.toLowerCase());
  }

/** 
   * Returns true if the given string is a stop word.
   * 
   * @param str the word to test
   * @return true if the word is a stopword
   */
  public static boolean isStopword(String str) {
    return m_Stopwords.is(str.toLowerCase());
  }

}
